################################################################################
# Angle                                                                        #
#                                                                              #
"""Analyse the molecule angle in a pore."""
################################################################################


import seaborn as sns
import matplotlib.pyplot as plt

import poreana.utils as utils


def bins_plot(data_link_angle, data_link_dens, pore_id="shape_00", intent="", is_mean=False, is_norm=False, kwargs={}):
    """This function plots the angle. If an intent is given instead, only a
    plot-function will be called. Available options for ``intent`` are

    * empty string - Create subplots for the density inside and outside the pore
    * **in** - Create plot for the density inside pore
    * **ex** - Create plot for the density outside pore

    Parameters
    ----------
    data_link_dens : string
        Link to density hdf5, obj or yml data file generated by the sample rountine
        :func:`poreana.sample.Sample.init_density`
    data_link_angle : string
        Link to angle data object generated by the sample routine
        :func:`poreana.sample.Sample.init_angle`
    intent : string, optional
        Intent for plotting
    is_mean : bool, optional
        True to plot mean values
    is_norm : bool, optional
        True to normalize x-axis
    kwargs: dict, optional
        Dictionary with plotting parameters (only with given intent)

    Returns
    -------
    mean : dictionary
        Mean value of the angle inside and outside the pore in nm
    """
    # Load data
    agl = utils.load(data_link_angle)
    dens = utils.load(data_link_dens)

    is_pore = "pore" in agl
    width = {}
    if is_pore:
        width["in"] = agl["data"][pore_id]["in_width"][:-1] #
    else:
        width["in"] = []
    width["ex"] = agl["data"]["ex_width"]

    if is_norm:
        for key, val in width.items():
            x_max = val[-1]
            width[key] = [x/x_max for x in val]

    areas = ["in", "ex"] if is_pore else ["ex"]

    # Divide angle by density in bins
    angle = {"ex": [], "in": []}
    angle["in"] = [agl["data"][pore_id]["in"][i]/dens["data"][pore_id]["in"][i] if dens["data"][pore_id]["in"][i] else 0 for i in range(len(agl["data"][pore_id]["in"]))]
    angle["ex"] = [agl["data"]["ex"][i]/dens["data"]["ex"][i] if dens["data"]["ex"][i] else 0 for i in range(len(agl["data"]["ex"]))] 

    # Calculate mean angle
    mean = {area: sum(angle[area])/len(angle[area]) for area in areas}

    # Full plot
    if not intent:
        plt.subplot(211)
        sns.lineplot(x=width["in"], y=angle["in"])
        if is_mean:
            sns.lineplot(x=width["in"], y=[mean["in"] for x in width["in"]])

        plt.xlim([0, width["in"][-1]])
        plt.xlabel("Distance from pore center (nm)")
        plt.ylabel(r"Angle (deg)")
        plt.legend(["Angle", "Mean"])

        plt.subplot(212)
        sns.lineplot(x=width["ex"], y=angle["ex"])
        if is_mean:
            sns.lineplot(x=width["ex"], y=[mean["ex"] for x in width["ex"]])

        plt.xlim([0, width["ex"][-1]])
        plt.xlabel("Distance from reservoir end (nm)")
        plt.ylabel(r"Angle (deg)")
        plt.legend(["Angle", "Mean"])

    # Intent plots
    else:
        if intent not in ["in", "ex"]:
            print("Invalid intent. Check documentation for available options.")
            return

        sns.lineplot(x=width[intent], y=angle[intent], **kwargs)
        plt.xlim([0, width[intent][-1]])

    return mean
